from typing import Iterable, List

from pyaltium.magicstrings import SchematicPinType, SchematicRecordType


def byte_arr_str(s: bytes) -> bytes:
    """Get a string encoded in a byte array."""
    size = s[0]
    name = s[1 : size + 1]
    return name, s[size + 1 :]


def pinstr_to_records(s: bytes) -> List[dict]:
    """Actually take a pin string and turn it into usable records."""
    records = []

    rotations = {0: 0, 1: 90, 2: 180, 3: 270}

    while True:
        try:
            if not len(s) > 10:
                break

            record = {"RECORD": SchematicRecordType.PIN}
            # Trim beginning of string
            s = s[17:]
            description, s = byte_arr_str(s)

            if description:
                record["Description"] = description

            record["PinType"] = SchematicPinType(int.from_bytes(s[1:2], "big"))

            rot_hide = int.from_bytes(s[2:3], "big")
            record["Rotation"] = rotations[rot_hide & 0x03]

            record["Hide_Designator"] = bool(rot_hide & 0x08)
            record["Hide_Name"] = bool(rot_hide & 0x10)
            record["PinLength"] = int.from_bytes(s[3:4], "big")
            record["Location.X"] = int.from_bytes(s[4:6], "big", signed=True)
            record["Location.Y"] = int.from_bytes(s[6:8], "big", signed=True)

            s = s[13:]
            record["Name"], s = byte_arr_str(s)
            record["Designator"], s = byte_arr_str(s)

            records.append(record)

            s = s[5:]

        except IndexError:
            pass

    return records


def handle_pin_records(records: Iterable[dict]) -> list:
    """Run through a list of records for a schematic component and handle pins.

    Pins are a bit weird. There is no record type for them so they are just binary
    strings in the middle of other textual records.

    That means we need to go through all the records and explicitely split this off,
    since they just tag along with whatever record preceeded them.


    |
    """

    retlist = []

    for rec in records:
        newrecords = []
        for key, val in rec.items():
            # If there is nothing bytes in the string, we are set
            if b"\x00" not in val:
                continue

            # Otherwise, do cleanup
            newval, pinstr = val.split(b"\x00", 1)
            rec[key] = newval

            # If it's too short to be a pin, it's probably just junk so ignore
            if len(pinstr) > 20:
                complete_str = b"\x00" + pinstr
                newrecords.extend(pinstr_to_records(complete_str))

        retlist.append(rec)
        retlist.extend(newrecords)

    return retlist



"""
types.py

Everything needed to interact with SchLib files
"""


from typing import Iterable, Union

import matplotlib.pyplot as plt
import olefile

from pyaltium.exceptions import PyAltiumError
from pyaltium.helpers import read_decode_stream
from pyaltium.magicstrings import MAX_READ_SIZE_BYTES


class OleMixin:
    """Helper functions for anything with an ole _file_name object."""

    def _list_storages(self):
        with olefile.OleFileIO(self._file_name) as ole:
            return ole.listdir()

    def _read_decode_stream(
        self,
        streamname: Union[str, Iterable],
        readbytes: int = MAX_READ_SIZE_BYTES,
        decode: str = "utf8",
    ) -> str:
        """Read a stream (in one go) and decode it. Maybe add yield in the future."""
        with olefile.OleFileIO(self._file_name) as ole:
            try:
                str_read = ole.openstream(streamname).read(readbytes)
                if decode:
                    return str_read.decode(decode, "ignore").removesuffix("\x00")
                return str_read
            except OSError:
                # Can't find stream
                return "" if decode else b""


class AltiumFileType(OleMixin):
    """This class will generally not be exposed.
    Just intended to set up children
    """

    def __init__(self, file_name: str = None) -> None:
        # Initialize variables to be used later
        self._file_name = None
        self._header_dict = {}
        self._section_keys_list = []
        self.items_list = []

        if file_name is not None:
            self.set_file_name(file_name)

    def __repr__(self):
        return self._file_name

    def set_file_name(self, file_name: str) -> None:
        """Check if file is valid (to the best of our ability) then update
        generic information.
        """
        self._file_name = file_name

        if not olefile.isOleFile(file_name):
            raise PyAltiumError("Unable to open file. Is it actually an Altium binary?")

        if not self._verify_file_type(file_name):
            raise PyAltiumError("Appears to be the wrong file type.")

        self._update_header_and_section_keys()
        self._update_item_list()

    def _update_header_and_section_keys(self) -> None:
        """Just update class's _header_dict object."""
        raise NotImplementedError()

    def _update_item_list(self) -> None:
        """Generate a list. This will happen in the inherited class."""
        raise NotImplementedError()

    def _verify_file_type(self, fname: str) -> bool:
        """Verify the file type is what is expected."""
        raise NotImplementedError()


class AltiumLibraryType(AltiumFileType):
    def list_items(self, as_dict=True) -> list:
        """Return a list of all the items.

        Optionally return them as a dictionary."""

        if not as_dict:
            return self.items_list

        return [item.as_dict() for item in self.items_list]


class AltiumLibraryItemType(OleMixin):
    def __init__(self) -> None:
        self._loaded_data = None
        self.name = None

    def _as_dict(self) -> dict:
        raise NotImplementedError

    def _run_load(self) -> None:
        """Set self._storage as needed."""
        raise NotImplementedError

    def _load(self) -> None:
        """Load data from the owner file."""
        if self._loaded_data is not None:
            return
        self._run_load()

    def _draw(self, ax: plt.Axes) -> None:
        raise NotImplementedError

    def get_svg(self):
        self._load()
        fig, ax = plt.subplots()
        ax.set_aspect("equal")
        self._draw(ax)
        ax.axis("off")
        ax.autoscale(tight=True)
        fig.savefig(
            f"testout/{self.name}.svg",
            dpi=1200,
            transparant=True,
            bbox_inches="tight",
            pad_inches=0,
        )
        fig.savefig(
            f"testout/{self.name}.png",
            dpi=1200,
            transparant=True,
            bbox_inches="tight",
            pad_inches=0,
        )



import math

import matplotlib.patches as patches
import matplotlib.pyplot as plt

from pyaltium.base import AltiumLibraryItemType, AltiumLibraryType
from pyaltium.helpers import (
    altium_string_split,
    altium_value_from_key,
    eval_bool,
    eval_color,
    getfloat,
    getint,
    normalize_dict,
    re_before_first_record,
    sch_sectionkeys_to_dict,
)
from pyaltium.magicstrings import SCHLIB_HEADER, SchematicRecordType, get_sch_record
from pyaltium.schematichelpers import handle_pin_records


class SchLib(AltiumLibraryType):
    """Main object to interact with schematic libraries."""

    def _verify_file_type(self, fname: str) -> bool:
        """Check if our magic string is in the header."""
        fh_str = self._read_decode_stream("FileHeader", 128)
        return SCHLIB_HEADER in fh_str

    def _update_header_and_section_keys(self) -> None:
        """Just update class's _header_dict object."""
        fh_str = self._read_decode_stream("FileHeader")
        sk_str = self._read_decode_stream("SectionKeys")

        self._header_dict = altium_string_split(fh_str)
        self._section_keys_list = altium_string_split(sk_str)

    def _update_item_list(self) -> None:
        """Override main class, just update the list of items in the library.

        Most of this information is kept in the file header. However, we need
        to get some information from sectionkeys if names got truncated in the header.
        """
        d = self._header_dict

        # Get the component count so we know what to look for
        item_count = int(altium_value_from_key(d, "CompCount"))

        self.items_list = []

        sec_keys = sch_sectionkeys_to_dict(self._section_keys_list)

        # Loop through each item listed in the fileheader
        for i in range(item_count):
            libref = altium_value_from_key(d, f"LibRef{i}")
            description = altium_value_from_key(d, f"CompDescr{i}")
            partcount = int(altium_value_from_key(d, f"PartCount{i}")) - 1

            if libref in sec_keys:
                sectionkey = sec_keys[libref]
            else:
                sectionkey = libref

            self.items_list.append(
                SchLibItem(
                    libref=libref,
                    description=description,
                    partcount=partcount,
                    sectionkey=sectionkey,
                    parent_fname=self._file_name,
                )
            )


class SchLibItem(AltiumLibraryItemType):
    def __init__(
        self,
        libref: str,
        sectionkey: str,
        description: str,
        partcount: int,
        parent_fname: str,
    ) -> None:
        super().__init__()
        self.libref = libref
        self.name = libref
        self.sectionkey = sectionkey
        self.description = description
        self.partcount = partcount
        self._file_name = parent_fname

    def _run_load(self) -> None:
        pin_text_data = self._read_decode_stream(
            (self.sectionkey, "PinTextData"), decode=False
        )
        data = self._read_decode_stream((self.sectionkey, "Data"), decode=False)

        # Remove everything before the first "|RECORD"
        while not data.startswith(b"|RECORD"):
            data = data[1:]

        # Split into records
        records = [b"|RECORD" + d for d in data.split(b"|RECORD")[1:]]

        # Split these into their parameters. We need to temporarily escape the
        # |&| that is sometimes used.
        records = [
            dict(
                split
                for s in rec.replace(b"|&|", b"&&&&").split(b"|")[1:]
                if len(split := s.replace(b"&&&&", b"|&|").split(b"=", 1)) == 2
            )
            for rec in records
        ]

        print(f"Processing {self.name}")
        records = handle_pin_records(records)

        pass
        # Result is something like
        # [{"RECORD": "34", "Location.X": "-5", "Location.Y": "10",...},...]

        self._loaded_data = [SchematicRecord(rec) for rec in records]

    def _draw(self, ax: plt.Axes) -> None:
        """Create the drawing on the axes"""
        records = self._loaded_data
        part_display_mode = 1
        for record in records:
            record.draw(ax, part_display_mode)

    def as_dict(self) -> dict:
        """Create a parsable dict."""
        return {
            "libref": self.libref,
            "description": self.description,
            "partcount": self.partcount,
            "sectionkey": self.sectionkey,
        }

    def __repr__(self) -> str:
        return f"<SchLibItem> {self.name}"


class SchematicRecord:
    """An object record stored in a schematic."""

    def __init__(self, parameters: dict) -> None:
        self.parameters = normalize_dict(parameters)
        self.record_type = get_sch_record(self.parameters.get("RECORD", 0))

    def draw(self, ax: plt.Axes, part_display_mode: int = 1) -> None:
        """Draw this single object on matplotlib axes."""
        typ = self.record_type
        params = self.parameters

        try:
            display_mode = int(params.get("OwnerPartDisplayMode", 1))
            part_id = params.get("OwnerPartID", 1)

            if display_mode != part_display_mode:
                return

            loc_x = getint(params, "Location.X")
            loc_y = getint(params, "Location.Y")
            rotation = getint(params, "Rotation")
            linewidth = getfloat(params, "LineWidth", 0.4) * 10
            color = eval_color(params.get("Color"))

            if typ == SchematicRecordType.RECTANGLE:
                tr_x = getint(params, "Corner.X")
                tr_y = getint(params, "Corner.Y")
                is_solid = eval_bool(params.get("IsSolid", "1"))
                fill_color = eval_color(params.get("AreaColor"))

                fill_color = fill_color if is_solid else "none"
                print(f"RECT {(loc_x, loc_y)} {(tr_x - loc_x,tr_y - loc_y)}")
                rect = patches.Rectangle(
                    (loc_x, loc_y),
                    width=tr_x - loc_x,
                    height=tr_y - loc_y,
                    linewidth=linewidth,
                    edgecolor=color,
                    facecolor=fill_color,
                )
                ax.add_patch(rect)

            elif typ == SchematicRecordType.PIN:
                pinlength = getint(params, "PinLength")
                print(f"PIN {(loc_x, loc_y)} {pinlength}")
                x1 = loc_x + math.cos(math.radians(rotation)) * pinlength
                y1 = loc_y + math.sin(math.radians(rotation)) * pinlength
                ax.plot((loc_x, x1), (loc_y, y1), "k", linewidth=linewidth)

            elif typ == SchematicRecordType.LABEL:
                just = getint(params, "Justification", 0)
                justMap = {
                    0: ("bottom", "left"),
                    1: ("bottom", "center"),
                    2: ("bottom", "right"),
                    3: ("center", "left"),
                    4: ("center", "center"),
                    5: ("center", "right"),
                    6: ("top", "left"),
                    7: ("top", "center"),
                    8: ("top", "right"),
                }
                jm = justMap[just]
                ax.text(
                    loc_x,
                    loc_y,
                    params.get("Text", ""),
                    color=color,
                    verticalalignment=jm[0],
                    horizontalalignment=jm[1],
                )

            # else:
            #     print(typ)

            # elif record.record_type==SchematicRecord.

        except KeyError:
            # If we are missing a key, we wouldn't be able to draw properly
            pass

    def __repr__(self) -> str:
        return f"<SchematicRecord> {self.record_type.name}"


# bytes(pins[0], encoding="raw_unicode_escape").hex()
# hx=[bytes(p, encoding="raw_unicode_escape").hex(' ',4) for p in pins]



"""helpers.py"""
import re
from typing import Any

import olefile

from pyaltium.magicstrings import MAX_READ_SIZE_BYTES

re_before_first_record = re.compile(r"^.*?(\|RECORD)")
re_cleanstr = re.compile(r"[^\w_\.]")

# Ignore |& bridges, like for pin records
re_split_exclude_ampersand = re.compile(r"\|(?<!\|&)")


def altium_string_split(s: str) -> list:
    """Just split into a list by the pipe character"""
    return s.split("|")


def altium_value_from_key(arr: list, key: str) -> str:
    """Loop through and replace "key" part."""
    for item in arr:
        if item.startswith(key):
            return item.replace(f"{key}=", "")

    # Just return empty if not found
    return ""


def sch_sectionkeys_to_dict(arr: list) -> dict:
    """
    Turn sectionkeys into a dict with the form:
    {
        libref1: sectionkey
        libref2: sectionkey
    }
    """
    key_count = altium_value_from_key(arr, "KeyCount")
    if key_count == "":
        key_count = 0
    else:
        key_count = int(key_count)

    retdict = {}

    for i in range(0, key_count - 1):
        retdict[altium_value_from_key(arr, f"LibRef{str(i)}")] = altium_value_from_key(
            arr, f"SectionKey{str(i)}"
        )

    return retdict


def read_decode_stream(
    filename: str, streamname: str, readbytes: int = MAX_READ_SIZE_BYTES
) -> str:
    """Read a stream (in one go) and decode it. Maybe add yield in the future."""
    with olefile.OleFileIO(filename) as ole:
        try:
            stream = ole.openstream(streamname)
            return stream.read(readbytes).decode("utf8")
        except OSError:
            # Can't find stream
            return ""


def eval_bool(b: str) -> bool:
    """Evaluate possible bool values"""
    return b.lower() in ("1", "t", "true")


def eval_color(c: str) -> str:
    """Fix the dumb color flip flop."""
    if c is None:
        return "#FFFFFF"
    ci = int(c)
    r = ci & 0x0000FF
    g = (ci & 0x00FF00) >> 8
    b = (ci & 0xFF0000) >> 16

    return f"#{r:02x}{g:02x}{b:02x}"


def normalize_dict(d: dict) -> dict:
    """Decode dictionary key/values and remove extra binary characters."""

    def norm_value(x: Any):
        if not isinstance(x, bytes):
            return x
        return re_cleanstr.sub("", x.decode("utf8", "ignore"))

    return {norm_value(k): norm_value(v) for k, v in d.items() if k and v}


def getfloat(params: dict, key: str, default: float = 0) -> float:
    return float(params.get(key, default))


def getint(params: dict, key: str, default: int = 0) -> int:
    return int(params.get(key, default))



import olefile

from pyaltium.base import AltiumLibraryItemType, AltiumLibraryType
from pyaltium.helpers import altium_string_split, altium_value_from_key
from pyaltium.magicstrings import MAX_READ_SIZE_BYTES, PCBLIB_HEADER


class PcbLib(AltiumLibraryType):
    """Main object to interact with PCBLib"""

    def _verify_file_type(self, fname: str) -> bool:
        """Check if our magic string is in the header"""
        fh_str = self._read_decode_stream("FileHeader", 128)
        return PCBLIB_HEADER in fh_str

    def _update_header_and_section_keys(self) -> None:
        """Just update class's _header_dict object"""
        fh_str = self._read_decode_stream("FileHeader")
        sk_str = self._read_decode_stream("SectionKeys")

        self._header_dict = altium_string_split(fh_str)
        self._section_keys_list = altium_string_split(sk_str)

    def _update_item_list(self) -> None:
        with olefile.OleFileIO(self._file_name) as ole:
            # Just list storages. We will need to add something to integrate
            # SectionKeys at some point, but the PCBLib flavor of that
            # file makes 0 sense (yet)
            storages_list = ole.listdir(streams=False, storages=True)

            self.items_list = []

            # Need to select only items in storages_list with len 1 (any more
            # would be a subdir) then select 0th element (to get list of str
            # rather than list of list of str)
            for lib_item in (s for s in storages_list if len(s) == 1):
                # Ignore this metadata stream
                if ("fileversioninfo" in lib_item[0].lower()) or (
                    "library" in lib_item[0].lower()
                ):
                    continue

                # We want the paramaters stream within our storage
                lib_item.append("Parameters")

                param_bytestring = ole.openstream(lib_item).read(MAX_READ_SIZE_BYTES)

                # First 4 bytes seem to be random noise
                param_bytestring = param_bytestring[4:]

                # Note: don't really want to ignore errors but
                # '3LED ArrayVertical 2mm TH' has a mystery character
                params_list = altium_string_split(
                    param_bytestring.decode("utf8", errors="ignore")
                )

                footprintref = altium_value_from_key(params_list, "PATTERN")
                description = altium_value_from_key(params_list, "DESCRIPTION")

                height_tmp = altium_value_from_key(params_list, "HEIGHT").lower()

                if "mm" in height_tmp:
                    height = round(float(height_tmp.replace("mm", "")), 2)

                if "mil" in height_tmp:
                    height = round(float(height_tmp.replace("mil", "")) * 0.0254, 2)

                self.items_list.append(
                    PcbLibItem(
                        footprintref=footprintref,
                        description=description,
                        height=height,
                        parent_fname=self._file_name,
                    )
                )


class PcbLibItem(AltiumLibraryItemType):
    def __init__(
        self,
        footprintref: str,
        description: str,
        height: float,
        parent_fname: str,
    ) -> None:
        super().__init__()
        self.footprintref = footprintref
        self.description = description
        self.height = height
        self._file_name = parent_fname

    def _run_load(self) -> None:
        raise NotImplementedError

    def as_dict(self) -> dict:
        """Create a parsable dict."""
        return {
            "footprintref": self.footprintref,
            "description": self.description,
            "height": self.height,
        }



"""magicstrings.py"""
from enum import IntEnum, unique
from typing import Union

SCHLIB_HEADER = (
    "HEADER=Protel for Windows - Schematic Library Editor Binary File Version 5.0"
)
PCBLIB_HEADER = "PCB 6.0 Binary Library File"
MAX_READ_SIZE_BYTES = 1024 * 1024


@unique
class SchematicRecordType(IntEnum):
    """These types are stored in a schematic file."""

    UNDEFINED = 0
    COMPONENT = 1
    PIN = 2
    IEEE_SYMBOL = 3
    LABEL = 4
    BEZIER = 5
    POLYLINE = 6
    POLYGON = 7
    ELLIPSE = 8
    PIECHART = 9
    RECTANGLE_ROUND = 10
    ELIPTICAL_ARC = 11
    ARC = 12
    LINE = 13
    RECTANGLE = 14
    SHEET_SYMBOL = 15
    SHEET_ENTRY = 16
    POWER_PORT = 17
    PORT = 18
    NO_ERC = 22
    NET_LABEL = 25
    BUS = 26
    WIRE = 27
    TEXT_FRAME = 28
    JUNCTION = 29
    IMAGE = 30
    SHEET = 31
    SHEET_NAME = 32
    FILE_NAME = 33
    DESIGNATOR = 34
    BUS_ENTRY = 37
    TEMPLATE = 39
    PARAMETER = 41
    IMPLEMENTATION_LIST = 44


@unique
class SchematicPinType(IntEnum):
    """Possible types for a pin."""

    INPUT = 0
    IO = 1
    OUTPUT = 2
    OPEN_COLLECTOR = 3
    PASSIVE = 4
    HIGH_Z = 5
    OPEN_EMITTER = 6
    POWER = 7


def get_sch_record(value: Union[str, int]):
    try:
        if isinstance(value, str):
            return SchematicRecordType(int(value))
        return SchematicRecordType(value)
    except ValueError:
        return SchematicRecordType(0)



import re

import setuptools

# from .src.pyaltium import __version__

with open("src/pyaltium/__init__.py") as f:
    txt = f.read()
    r = re.search(r'__version__\s+=\s+"(.*)"', txt)
    __version__ = r.group(1)

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="PyAltium",
    version=__version__,
    author="Trevor Gross",
    author_email="t.gross35@gmail.com",
    description="A package for reading Altium files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tgross35/PyAltium",
    project_urls={
        "Bug Tracker": "https://github.com/tgross35/PyAltium/issues",
    },
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    python_requires=">=3.7",
)
